"""CSDL generation library."""

import os.path

import yaml

__version__ = '0.2.0'
__author__ = 'Sam Clements <sam@borntyping.co.uk>'
__all__ = (
    'Condition', 'Definition', 'InvalidCSDL', 'Return', 'ScoredTag',
    'Statement', 'Tag', 'Target', 'TargetAttribute', 'TargetOperator',
    'interaction', 'scores', 'tags', 'twitter'
)

TARGETS_DIR = os.path.dirname(os.path.realpath(__file__))
TARGETS_PATH = os.path.join(TARGETS_DIR, 'targets.yaml')

with open(TARGETS_PATH, 'r') as f:
    TARGETS = yaml.safe_load(f)


class InvalidCSDL(Exception):
    pass


class Statement(str):
    pass


class Condition(Statement, str):
    def __and__(self, other):
        return Statement(self + " AND " + other)

    def __or__(self, other):
        return Statement(self + " OR " + other)

    def not_(self):
        return Condition(" NOT " + self)


class Return(object):
    def __init__(self, csdl):
        self.csdl = csdl

    def __str__(self):
        return "return { " + self.csdl + " }"


class Tag(object):
    def __init__(self, name, statement):
        self.name = name
        self.statement = statement

    def __str__(self):
        return "tag \"{self.name}\" {{ {self.statement} }}".format(self=self)


class ScoredTag(Tag):
    def __init__(self, name, statement, score):
        super(ScoredTag, self).__init__(name, statement)
        self.score = score

    def __str__(self):
        return "tag.{name} {score} {{ {statement} }}".format(**self.__dict__)


class Definition(object):
    def __init__(self, return_statement=None):
        self.statements = []

        if return_statement:
            self.returns(return_statement)

    def __str__(self):
        return "\n".join(map(str, self.statements))

    def prepend(self, statement):
        """Insert a statement before the final statement."""
        self.insert((len(self) - 1) or 0, statement)

    def __lshift__(self, other):
        return self.statements.append(other)

    def tag(self, name, csdl):
        self.statements.prepend(Tag(name, csdl))

    def returns(self, return_statement):
        for statement in self.statements:
            if isinstance(statement, Return):
                raise InvalidCSDL("Cannot have two return statements")
        self.statements.append(Return(return_statement))


class TargetAttribute(object):
    def __init__(self, _name):
        self._name = _name

    def __get__(self, instance, owner):
        return Target(instance._name + "." + self._name)


class TargetOperator(object):
    def __init__(self, operator, accepts=None, cs=False):
        self.operator = operator
        self.accepts = accepts
        self.cs = cs

    def __get__(self, instance, owner):
        """Create an operator function."""
        if self.cs:
            def operator(argument=None, cs=False):
                return self.condition(instance, argument, cs)
        else:
            def operator(argument=None):
                return self.condition(instance, argument)
        return operator

    def condition(self, target, argument=None, cs=False):
        if cs:
            segments = [target, "cs", self.operator]
        else:
            segments = [target, self.operator]

        if self.accepts is not None:
            # Check that the argument is acceptable
            if not isinstance(argument, self.accepts):
                raise InvalidCSDL("Argument must be {0}".format(self.accepts))

            # Deal with lists
            if isinstance(argument, (list, tuple)):
                argument = ", ".join(argument)

            # String arguments are wrapped with quotes
            if isinstance(argument, str):
                argument = '"' + argument + '"'

            segments.append(argument)

        return Condition(" ".join(map(str, segments)))


class Target(object):
    @classmethod
    def create(cls, name, data, parent=None):
        attributes = dict()

        for item in data.pop('attributes', []):
            attributes[item] = TargetAttribute(item)

        for key, value in data.pop('subclasses', {}).items():
            attributes[key.lower()] = Target.create(key, value, parent=name)

        # Use the parent's name as a prefix if this is a child target
        clsname = name if parent is None else parent + name

        cls = type(clsname, (cls,), attributes)

        return cls(name.lower())

    def __init__(self, _name):
        self._name = _name

    def __repr__(self):
        return "<{} '{}'>".format(self.__class__.__name__, self._name)

    def __str__(self):
        return self._name

    def __get__(self, instance, owner):
        """Used by subattributes to return self prefixed by the parent."""
        return self.__class__(instance._name + "." + self._name)

    # Operators

    contains = TargetOperator("contains", accepts=str, cs=True)

    substr = TargetOperator("substr", accepts=str, cs=True)

    contains_any = TargetOperator("contains_any", accepts=list, cs=True)
    any = contains_any

    contains_all = TargetOperator("contains_all", accepts=list, cs=True)
    all = contains_all

    exists = TargetOperator("exists")

    _in = TargetOperator("in", accepts=list, cs=True)

    url_in = TargetOperator("url_in", accepts=list, cs=True)

    __eq__ = TargetOperator("==", accepts=(int, float, str), cs=True)
    equals = __eq__

    __ne__ = TargetOperator("!=", accepts=(int, float, str), cs=True)
    notequal = __ne__

    regex_partial = TargetOperator("regex_partial", accepts=str)

    regex_exact = TargetOperator("regex_exact", accepts=str)

# Interaction is the parent of all others
interaction = Target.create("interaction", TARGETS.pop('Interaction'))

for _name, _attributes in TARGETS.items():
    globals()[_name.lower()] = interaction.create(_name, _attributes)


def tags(
        tags,
        tag_csdl=interaction.content.contains,
        returns_csdl=interaction.content.contains_any):
    """Create a simple CSDL tag filter."""
    result = Definition()

    for tag in tags:
        result << Tag(tag, tag_csdl(tag))

    if returns_csdl is not None:
        result.returns(returns_csdl(tags))

    return result


def scores(
        tags,
        match_csdl=interaction.content.contains,
        returns_csdl=interaction.content.contains_any):
    """Create a simple CSDL scores filter."""
    result = Definition()

    for tag, score in tags.items():
        result << ScoredTag(tag, match_csdl(tag), score)

    if returns_csdl is not None:
        result.returns(returns_csdl(tags.keys()))

    return result
